(* pa_rascl.ml -- Copyright  2006 by Matthew C. Gushee 
   This program is free software, released under the terms of the BSD
   license. See the file LICENSE for details. *)

(* ================================================================== *)
(* --  DECLARATIONS  ------------------------------------------------ *)
open Rascl
open Dict

type type_tag = [ `S | `I | `F | `B ]
(*
type example_data =
  [ `S of string * string * string option
  | `L of string * string * string list option 
  | `D of string * example_data Queue.t
  | `C of example_data Queue.t ]
*)
type example_data =
  [ `S of string * type_tag * string option
  | `L of string * type_tag * string list option 
  | `D of string * example_data Queue.t ]
type simple_value =
  | Ss of string option
  | Sl of string list option
type intermediate_data = {
  set_expr : MLast.expr option;
  doin_expr : MLast.expr option;
  getter : MLast.class_str_item;
  setter : MLast.class_str_item option;
  sub_obj : MLast.class_str_item option;
  example : example_data;
}
(* ================================================================== *)


(* ================================================================== *)
(* --  UTILITY STRUCTURES AND FUNCTIONS  ---------------------------- *)
let push_path, pop_path =
  let (pstack:string Stack.t) = Stack.create () in
  ( fun p -> Stack.push p pstack ),
  ( fun () ->
    let path = ref [Stack.pop pstack] in
    Stack.iter (fun elt -> path := elt::!path) pstack;
    !path )

let def_repr defval =
  match defval with
  | Ss (Some s) -> "\"" ^ s ^ "\""
  | Ss None -> "?" 
  | Sl (Some l) ->
    ( match l with
      | [] -> "[]"
      | hd::tl ->
        let pfx = "[" ^ hd in
        ( List.fold_left (fun str elt -> str ^ "; " ^ elt) hd tl ) ^ "]" )
  | Sl None -> "?" 

let exenq decls =
  let q = Queue.create () in
  List.iter
    ( fun d -> Queue.push d.example q )
    decls;
  q

let write_example decls =
  let ex_file =
    try Sys.getenv "SC_EXAMPLE"
    with Not_found -> "config.example" in
  let oc = open_out ex_file in
  let friendly stag ttag =
    match stag, ttag with
    | `S, `S -> "string"
    | `S, `I -> "integer" 
    | `S, `F -> "float" 
    | `S, `B -> "boolean" 
    | `L, `S -> "string list"
    | `L, `I -> "integer list" 
    | `L, `F -> "float list" 
    | `L, `B -> "boolean list" 
    | `D, _  -> "dictionary"
  and defstr def =
    match def with
    | None -> "?"
    | Some d -> d in
  let write_string_string ?(indent="") oc key default =
    output_string oc
      (Printf.sprintf "%s%s : \"%s\"  # string\n" indent key (defstr default));
    flush stdout
  and write_string ?(indent="") oc key typ default =
    output_string oc
      (Printf.sprintf "%s%s : %s   # %s\n" indent key (defstr default) (friendly `S typ));
    flush stdout in
  let string_of_list ?(quoted=false) lst =
    let repr s = if quoted then "\"" ^ s ^ "\"" else s in
    let rec build_string ?(init=false) lst' str =
      match lst', init with
      | [], true -> "[]"
      | [], false -> str ^ "]"
      | h::t, true -> build_string t ("[" ^ (repr h))
      | h::t, false -> build_string t (str ^ ", " ^ (repr h)) in
    match lst with
    | None -> "[]"
    | Some l -> build_string ~init:true l "" in
  let write_string_list ?(indent="") oc key default =
    output_string oc
      ( Printf.sprintf "%s%s : %s  # string list\n"
        indent key (string_of_list ~quoted:true default) );
    flush stdout
  and write_list ?(indent="") oc key typ default =
    output_string oc
      ( Printf.sprintf "%s%s : %s  # %s\n"
        indent key (string_of_list default) (friendly `L typ) );
    flush stdout in
  let rec write_top decls' =
    match decls' with
    | [] -> ()
    | d::dd ->
      ( match d.example with
        | `S (k, `S, d) -> write_string_string oc k d 
        | `S (k, t, d) -> write_string oc k t d
        | `L (k, `S, d) -> write_string_list oc k d
        | `L (k, t, d) -> write_list oc k t d
        | `D (k, q) ->
          output_string oc (Printf.sprintf "%s : {\n" k); flush stdout;
          write_sub q "  ";
          output_string oc "}\n"; flush stdout );
      write_top dd
  and write_sub q indent =
    Queue.iter
      ( fun elt ->
        match elt with
        | `S (k, `S, d) -> write_string_string ~indent:indent oc k d
        | `S (k, t, d) -> write_string ~indent:indent oc k t d
        | `L (k, `S, d) -> write_string_list ~indent:indent oc k d
        | `L (k, t, d) -> write_list ~indent:indent oc k t d
        | `D (k, q) ->
          output_string oc (Printf.sprintf "%s%s : {\n" indent k); flush stdout;
          write_sub q (indent ^ "  ");
          output_string oc (Printf.sprintf "%s}\n" indent); flush stdout )
    q in
  write_top decls;
  close_out oc
    

let string_list_node values =
  let loc = Token.dummy_loc in
  List.fold_right
    ( fun elt lst -> <:expr< [ $str:elt$ :: $lst$ ] >> )
    values <:expr< [] >>

let expr_list_node values =
  let loc = Token.dummy_loc in
  List.fold_right
    ( fun elt lst -> <:expr< [ $elt$ :: $lst$ ] >> )
    values <:expr< [] >>
(* ================================================================== *)


(* ================================================================== *)
(* --  DEFAULTS  ---------------------------------------------------- *)
let string_set_expr loc key value =
  match value with
  | None -> None
  | Some v ->
    Some <:expr< set $str:key$ (S $str:v$) >>

let list_set_expr loc key value =
  match value with
  | None -> None
  | Some v ->
    let lst = string_list_node v in
    Some <:expr< set $str:key$ (L $lst$) >>

let subdict_set_expr loc key =
  Some <:expr< set $str:key$ (D (create ())) >>

let doin_expr_body loc path decls =
  let actions =
    List.fold_right
      ( fun d lst ->
        match d.set_expr, d.doin_expr with
        | Some se, Some de -> se :: de :: lst
        | Some se, None -> se :: lst
        | None, Some de -> de :: lst
        | None, None -> lst )
      decls [] in
  string_list_node path,
  expr_list_node actions

let doin_expr loc path decls =
  let pnode, anode = doin_expr_body loc path decls in
  Some <:expr< do_in $pnode$ $anode$ >>

let top_doin_expr loc path decls =
  let pnode, anode = doin_expr_body loc path decls in
  <:expr< do_in $pnode$ $anode$ defaults >>
(* ================================================================== *)


(* ================================================================== *)
(* --  ACCESSORS  --------------------------------------------------- *)
let string_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_string $str:key$ >>
let string_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_string $str:key$ v >>

let int_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_int $str:key$ >>
let int_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_int $str:key$ v >>

let float_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_float $str:key$ >>
let float_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_float $str:key$ v >>

let bool_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_bool $str:key$ >>
let bool_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_bool $str:key$ v >>

let string_list_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_string_list $str:key$ >>
let string_list_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_string_list $str:key$ v >>

let int_list_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_int_list $str:key$ >>
let int_list_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_int_list $str:key$ v >>

let float_list_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_float_list $str:key$ >>
let float_list_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_float_list $str:key$ v >>

let bool_list_getter_csi loc key =
  <:class_str_item< method $lid:key$ = self#get_bool_list $str:key$ >>
let bool_list_setter_csi loc key =
  let mname = "set_" ^ key in
  Some <:class_str_item< method $mname$ v = self#set_bool_list $str:key$ v >>

let subconf_getter_csi loc key =
  let oname = key ^ "_" in
  <:class_str_item< method $lid:key$ = $lid:oname$ >> 
(* ================================================================== *)


(* ================================================================== *)
(* --  OBJECTS  ----------------------------------------------------- *)
let object_body loc decls =
  let sub_objects =
    let sos =
      List.fold_right
        ( fun idat lst ->
          match idat.sub_obj with
          | None -> lst
          | Some so -> so::lst ) 
        decls [] in
    <:class_str_item< declare $list:sos$ end >>
  and accessors =
    let accs =
      List.fold_right
        ( fun idat lst ->
          match idat.setter with
          | None -> idat.getter :: lst
          | Some se -> idat.getter :: se :: lst )
        decls [] in
    <:class_str_item< declare $list:accs$ end >> in
  (sub_objects, accessors)

let subconf_csi loc key decls =
  let sub_objects, accessors = object_body loc decls
  and pself = <:patt< self >>
  and inheritance =
    <:class_str_item< 
      inherit sub_config data defaults path as super
    >> in
  let obj_expr =
    MLast.ExObj
      (loc, Some pself, [inheritance; sub_objects; accessors]) in
  let keylist_expr = <:expr< [$str:key$] >> in
  let path_bind_expr =
    <:expr<
      let path = path @ $keylist_expr$ in $obj_expr$
    >>
  and oname = key ^ "_" in
  Some <:class_str_item< value $lid:oname$ = $path_bind_expr$ >>

let rootconf_si loc cname decls =
  let sub_objects, accessors = object_body loc decls
  and pself = <:patt< self >>
  and inheritance =
    <:class_str_item<
      inherit root_config srcs dest data defaults as super
    >> in
  let oe =
    MLast.ExObj
      (loc, Some pself, [inheritance; sub_objects; accessors]) in
  <:str_item<
      value $lid:cname$ srcs dest =
        let data = Dict.create ()
        and path = [] in $oe$
  >>

let main_si loc cname decls = 
  write_example decls;
  let os = <:str_item< open Rascl >>
  and oc = <:str_item< open ConfigObject >>
  and od = <:str_item< open Dict >>
  and defcreate = <:str_item< value defaults = create () >>
  and defsetup0 = top_doin_expr loc [] decls in
  let defsetup = <:str_item< $exp:defsetup0$ >>
  and cl = rootconf_si loc cname decls in
  let all = [os; oc; od; defcreate; defsetup; cl] in
  <:str_item< declare $list:all$ end >>

(* ================================================================== *)


(* ================================================================== *)
(* --  GRAMMAR  ----------------------------------------------------- *)
EXTEND
  GLOBAL: Pcaml.str_item;
  Pcaml.str_item: FIRST [
    [ cname = LIDENT; "="; "{"; decls = LIST1 config_decl; "}"; OPT ";" ->
      main_si loc cname decls ]
  ];

  config_decl: [
    [ key = LIDENT; "="; "{"; decls = LIST1 config_decl; "}"; OPT ";" ->
        { set_expr = subdict_set_expr loc key;
          getter = subconf_getter_csi loc key;
          setter = None;
          sub_obj = subconf_csi loc key decls;
          doin_expr = doin_expr loc [key] decls;
          example = `D (key, (exenq decls)) }
    | key = LIDENT; ":"; "string"; "list";
        default = OPT string_list; ";" -> 
        (* let estr =
          key ^ " : " ^ (def_repr (Sl default)) ^ "  # string list" in *)
        { set_expr = list_set_expr loc key default;
          getter = string_list_getter_csi loc key;
          setter = string_list_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `L (key, `S, default) }
    | key = LIDENT; ":"; "int"; "list";
        default = OPT int_list; ";" -> 
        let estr =
          key ^ " : " ^ (def_repr (Sl default)) ^ "  # integer list" in
        { set_expr = list_set_expr loc key default;
          getter = int_list_getter_csi loc key;
          setter = int_list_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `L (key, `I, default) }
    | key = LIDENT; ":"; "float"; "list";
        default = OPT float_list; ";" -> 
        let estr =
          key ^ " : " ^ (def_repr (Sl default)) ^ "  # float list" in
        { set_expr = list_set_expr loc key default;
          getter = float_list_getter_csi loc key;
          setter = float_list_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `L (key, `F, default) }
    | key = LIDENT; ":"; "bool"; "list";
        default = OPT bool_list; ";" -> 
        let estr =
          key ^ " : " ^ (def_repr (Sl default)) ^ "  # bool list" in
        { set_expr = list_set_expr loc key default;
          getter = bool_list_getter_csi loc key;
          setter = bool_list_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `L (key, `B, default) }
    | key = LIDENT; ":"; "string";
      default = OPT [ "="; value = any_string -> value ]; ";" ->
        let estr =
          key ^ " : " ^ (def_repr (Ss default)) ^ "  # string" in
        { set_expr = string_set_expr loc key default;
          getter = string_getter_csi loc key;
          setter = string_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `S, default) }
    | key = LIDENT; ":"; "int";
      default = OPT [ "="; value = INT -> value ]; ";" ->
        let estr =
          key ^ " : " ^ (def_repr (Ss default)) ^ "  # integer" in
        { set_expr = string_set_expr loc key default;
          getter = int_getter_csi loc key;
          setter = int_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `I, default) }
    | key = LIDENT; ":"; "float";
      default = OPT [ "="; value = FLOAT -> value ]; ";" ->
        let estr =
          key ^ " : " ^ (def_repr (Ss default)) ^ "  # float" in
        { set_expr = string_set_expr loc key default;
          getter = float_getter_csi loc key;
          setter = float_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `F, default) }
    | key = LIDENT; ":"; "bool"; "="; "true"; ";" ->
        let estr =
          key ^ " : true  # boolean" in
        { set_expr = string_set_expr loc key (Some "true");
          getter = bool_getter_csi loc key;
          setter = bool_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `B, Some "true") }
    | key = LIDENT; ":"; "bool"; "="; "false"; ";" ->
        let estr =
          key ^ " : false  # boolean" in
        { set_expr = string_set_expr loc key (Some "false");
          getter = bool_getter_csi loc key;
          setter = bool_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `B, Some "false") }
    | key = LIDENT; ":"; "bool"; ";" ->
        let estr =
          key ^ " : ?  # boolean" in
        { set_expr = None;
          getter = bool_getter_csi loc key;
          setter = bool_setter_csi loc key;
          sub_obj = None;
          doin_expr = None;
          example = `S (key, `B, None) } ]
  ];
  string_list : [
    [ "="; "["; "]" -> []
    | "="; "["; s = any_string;
      l = LIST0 [","; s0 = any_string -> s0]; "]" -> s::l ]
  ];
  int_list : [
    [ "="; "["; "]" -> []
    | "="; "["; i = INT; l = LIST0 [","; i0 = INT -> i0]; "]" -> i::l ]
  ];
  float_list : [
    [ "="; "["; "]" -> []
    | "="; "["; f = FLOAT; l = LIST0 [","; f0 = FLOAT -> f0]; "]" -> f::l ]
  ];
  bool_list : [
    [ "="; "["; "]" -> []
    | "="; "["; b = boolean; l = LIST0 [","; b0 = boolean -> b0]; "]" -> b::l ]
  ];
  any_string : [
    [ s = STRING -> s | u = UIDENT -> u | l = LIDENT -> l ]
  ];
  boolean : [
    [ b = "true" -> b | b = "false" -> b ]
  ];
END;;
(* ================================================================== *)
