(* rascl.mli -- Copyright  2006 by Matthew C. Gushee 
   This program is free software, released under the terms of the BSD
   license. See the file LICENSE for details. *)

(** RASCL's A Simple Configuration Language. {b Rascl} is an OCaml library
    for processing RASCL files.

    {2 Introduction}

    This document describes the OCaml Rascl API. For a description of
    RASCL itself, please see the RASCL language documentation. You should
    have received that documentation with this package; the latest version
    can be found at
    {{:http://matt.gushee.net/rascl/}http://matt.gushee.net/rascl/}.

    This library, like the language, follows a minimalist philosophy. It
    makes no attempt at universality, but tries to provide a simple 
    interface with a few very widely applicable features. Thus there are
    a number of features that, while potentially useful, are deliberately
    left out. In particular, Rascl does not deal with the locations of
    configuration files. It recognizes only a few simple data types, and
    does not support any form of scripting.
    
    On the other hand, RASCL files should be easy to edit, even for
    people with minimal technical skills, and the Rascl API should be
    extremely simple to use. 

    There are two principal ways to access RASCL configuration data: the
    low level dictionary structure implemented in the [Dict] module, and
    a configuration object system, the foundation for which is provided
    in the [ConfigObject] module. Each has its advantages and disadvantages.

    The object system provides a very concise syntax. For example,
    a 'color' property might be accessed as:
    {[
      config_object#color
    ]}
  
    Properties are also type-safe: each accessor method returns (or accepts,
    for [set_*] methods) one of the RASCL data types, as specified in the
    configuration template that you create.
    
    On the other hand, using the object system requires that all property
    names be valid OCaml method names. There is also a potential problem 
    with configuration files containing values of incorrect types. Such 
    errors will not be discovered until the program attempts to retrieve
    the property from the object (as opposed to, say, failing when the file
    is loaded). At that point an exception will be raised and the property
    will be inaccessible until the file is corrected and reloaded.

    The dictionary structure is lower-level and less convenient. To access
    the 'color' property as in the above example, you would write:
    {[
      Dict.get "color" config_dict
    ]}
    Another important difference is that dictionaries only provide access
    to data in the form of strings (or lists thereof), though this will
    likely change in the near future.
    
    In exchange for less convenience, however, you gain significant
    flexibility. Dictionary keys, for example, can be arbitrary strings.
    You can also add and remove nested dictionaries, and even create
    new properties on the fly. 

    For further information, see the documentation for the [Dict] and
    [ConfigObject] modules.
    *)


(** This module implements the fundamental data structure of RASCL: the
    dictionary. This is a key-value structure whose keys are strings and
    whose values may be strings, lists, or dictionaries. Dictionaries
    have no fixed size, and their elements are mutable.  *)
module Dict : sig
  (** The type of dictionaries. *)
  type t

  (** The type for dictionary values. *)
  type data = S of string | L of string list | D of t

  val create : ?locked:bool -> unit -> t
    (** [Dict.create ?locked ()] creates a new, empty dictionary. The
        [locked] parameter indicates whether new nested dictionaries
        may be added. Its default value is [false], i.e. new nested
        dictionaries are allowed. *)
  val get : ?prefix:string list -> string -> t -> data
    (** [Dict.get ?prefix key dictionary] retrieves the value associated
        with [key] in [dictionary], raising [Not_found] if the key is not
        present. When retrieving an item from a subdictionary, [?prefix]
        may be used to specify the subdictionary. *)
  val set : ?prefix:string list -> string -> data -> t -> unit 
    (** [Dict.set ?prefix key value dictionary] sets the value for [key]
        in [dictionary] to [value]. When setting a value in a subdictionary,
        [?prefix] may be used to specify the subdictionary. *)
  val get_string : ?prefix:string list -> string -> t -> string
    (** [Dict.get_string ?prefix key dictionary] retrieves the value
        associated with [key] as a string. The stored value should match
        the pattern [S s], otherwise an exception will be raised.
        [?prefix] behaves as in [Dict.get]. *) 
  val get_dict : ?prefix:string list -> string -> t -> t
    (** [Dict.get_dict ?prefix key dictionary] retrieves the value
        associated with [key] as a dictionary. The stored value should
        match the pattern [D d], otherwise an exception will be raised.
        [?prefix] behaves as in [Dict.get]. *) 
  val get_list : ?prefix:string list -> string -> t -> string list
    (** [Dict.get_list ?prefix key dictionary] retrieves the value
        associated with [key] as a list. The stored value should match
        the pattern [L l], otherwise an exception will be raised.
        [?prefix] behaves as in [Dict.get]. *) 
  val set_string : ?prefix:string list -> string -> string -> t -> unit
    (** [Dict.set_string ?prefix key value dictionary] sets the value
        for [key] in [dictionary] to [S value]. [?prefix] behaves as in
        [Dict.set]. *)
  val set_list : ?prefix:string list -> string -> string list -> t -> unit
    (** [Dict.set_list ?prefix key value dictionary] sets the value
        for [key] in [dictionary] to [L value]. [?prefix] behaves as in
        [Dict.set]. *)
  val set_dict : ?prefix:string list -> string -> t -> t -> unit
    (** [Dict.set_dict ?prefix key value dictionary] sets the value
        for [key] in [dictionary] to [D value]. [?prefix] behaves as in
        [Dict.set]. *)
  val copy : t -> t
    (** [Dict.copy dictionary] returns a new dictionary whose elements
        are copies of those in the original dictionary. This is a
        deep-copy operation. *)


  (** {2 Iterators and filters}

      The following functions are similar to the [iter], [map], and
      [filter] functions associated with other data structures. Note,
      however, that there is no automatic recursion into subdictionaries.
      If you require such recursion, you must supply an appropriate
      recursive function. For example: {[

      let rec f key value =
        match value with
        | S s -> ....
        | L l -> ....
        | D subdict -> Dict.iter f subdict
  ]} *)
       
  val iter : (string -> data -> unit) -> t -> unit
    (** [Dict.iter f dictionary] applies [f] in turn to each element of
        [dictionary]. *)
  val map : (string -> data -> data) -> t -> t
    (** [Dict.map f dictionary] applies [f] in turn to each element of
        [dictionary], returning a new dictionary whose values are the
        results of [f] applied to each value in the original dictionary. *)
  val filter : (string -> data -> bool) -> t -> t
    (** [Dict.filter f dictionary] applies [f] in turn to each element of
        [dictionary], returning a new dictionary whose elements are copies
        of all those for which [f] returned true. *)
  val do_in : string list -> (t -> unit) list -> t -> unit
    (** [Dict.do_in context ff dictionary] executes in turn a list of
        functions, each of which takes a dictionary as its argument.
        [context] specifies the subdictionary to which [ff] should be
        applied; if [context] is an empty list, [ff] are applied to 
        [dictionary]. *)
  val get_from : string list -> string list -> t -> (string * data) list
    (** [Dict.get_from context keys dictionary] retrieves an association
        list of keys and values from [dictionary], or from a subdictionary
        identified by [context]. *)
  val set_in : string list -> (string * data) list -> t -> unit
    (** [Dict.set_in context elements dictionary] sets the key-value
        pairs specified by [elements] in dictionary, or in a subdictionary
        identified by [context]. *)

  (** The contents of this module are subject to change without notice.
      Use at your own risk. *)
  module Unsafe : sig
    val set_size : int -> unit
    val get_subdict : string list -> t -> t
    val detach_subdict : string list -> t -> t
    val alist_of_dict : ?strict:bool -> ?prefix:string list ->
      t -> (string * string) list
    val exists : ?prefix:string list -> (string -> data -> bool) ->
      t -> bool
    val lock_structure : t -> unit
    val unlock_structure : t -> unit
    val update_from : t -> t -> unit
    val dump : t -> unit
  end
end


(** An ocamlyacc-based parser for RASCL files. *)
module Parser : sig
  type token
  val config_dict : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> Dict.t
    (** [config_dict lexfun lexbuf] returns a RASCL dictionary. The only
        available implementation for [lexfun] is [Lexer.dict]. In general, 
        there is no need to use this function directly. Use
        [read_config_file] instead. *)
end


(** An ocamllex-based lexer for RASCL files. *)
module Lexer : sig
  exception LexerError of string
  val dict : Lexing.lexbuf -> Parser.token
end


(** This is a framework for the object-oriented interface to RASCL
    configurations. See the tutorial below for usage instructions. *)
module ConfigObject : sig
  (* class virtual config_base data defaults prefix = object(self) *)
  class virtual config_base : Dict.t -> Dict.t -> string list option -> object
    (**/**)
    method private get : string -> Dict.data
    method private set : string -> Dict.data -> unit
    method private get_string : string -> string
    method private get_bool : string -> bool
    method private get_int : string -> int
    method private get_float : string -> float
    method private get_list : string -> string list
    method private get_dict : string -> Dict.t
    method private get_string_list : string -> string list
    method private get_bool_list : string -> bool list 
    method private get_int_list : string -> int list
    method private get_float_list : string -> float list
    method private set_string : string -> string -> unit
    method private set_bool : string -> bool -> unit
    method private set_int : string -> int -> unit
    method private set_float : string -> float -> unit
    method private set_list : string -> string list -> unit
    method private set_string_list : string -> string list -> unit
    method private set_bool_list : string -> bool list -> unit
    method private set_int_list : string -> int list -> unit
    method private set_float_list : string -> float list -> unit
    method dump : unit
  end
  
  (* class virtual root_config (srcs:string list) (dest:string) data defaults = *)
  class virtual root_config :
      string list -> string -> Dict.t -> Dict.t -> object
    inherit config_base
    method load : unit -> unit
      (** Load (or reload) the configuration files specified when the
          object was created. I/O or permission errors are not handled. *)
    method dump : unit
      (** Print a textual representation of the object. Primarily
          useful for debugging. *)
  end
  
  (* class virtual sub_config data defaults prefix = object(self) *)
  class virtual sub_config : Dict.t -> Dict.t -> string list -> object
    inherit config_base
    method dump : unit
      (** Print a textual representation of the object. Primarily
          useful for debugging. *)
  end

(** {2 Tutorial}

    In order to use the OO interface, you need to create an object
    specific to your application. This can be done by using [pa_rascl.cmo],
    a CamlP4-based code generator. [pa_rascl.cmo] parses a {b configuration
    template} that you create and outputs a module containing a single
    function which creates a configuration object.

    {3 Templates}
    
    The configuration template syntax is as follows:
    {[
    template         : config-name space* '=' space* dictionary
    config-name      : ocaml-lowercase-ident
    dictionary       : '{' (property? | property (separator property)* ) '}'
    property         : name space* ':' space* type-and-default
                     | name space* '=' space* dictionary
    name             : ocaml-method-name
    type-and-default : 'string' (space* '=' space* string-expr)?
                     | 'int' (space* '=' space* int-expr)?
                     | 'float' (space* '=' space* float-expr)?
                     | 'bool' (space* '=' space* bool-expr)?
                     | 'string list' (space* '=' space* string-list-expr)?
                     | 'int list' (space* '=' space* int-list-expr)?
                     | 'float list' (space* '=' space* float-list-expr)?
                     | 'bool list' (space* '=' space* bool-list-expr)?
    ]}

    The prefix [ocaml-] indicates that the item conforms to OCaml
    syntax, and is fully documented in the OCaml manual (with the same
    name, minus 'ocaml-'.

    Items with the suffix [-expr] are documented in the RASCL language
    specification. In general, the syntax for these types follows that
    of OCaml, with the following exceptions:

    - Booleans are case-insensitive
    - The list separator is the comma (',') rather than the semicolon

    Default values are optional, but should be provided whenever possible;
    if an item has no default, a value {i must} be loaded from a file
    before attempting to access it at runtime.

    Property names must be unique within any given dictionary, but may be
    repeated in sub-dictionaries.

    For a sample template, see [food.ml]. You may also find it useful to
    compare this file with the supplied [config.example], which is the
    RASCL configuration generated from it.

    {3 Using the code generator}

    [pa_rascl.cmo] is a CamlP4 preprocessor. It depends on [pa_extend.cmo]
    and [q_MLast.cmo]. It can be invoked as follows:
    {[
      camlp4o -I <rascl_path> pa_extend.cmo q_MLast.cmo pa_rascl.cmo
    ]}
    So, the command to byte-compile a template called 'prefs.ml', using
    ocamlfind, might look like this:
    {[
      ocamlfind ocamlc -package rascl -pp \
        'camlp4o -I `ocamlfind query rascl` pa_extend.cmo q_MLast.cmo \
        pa_rascl.cmo' -c prefs.ml
    ]}
    The above command will produce two output files: the configuration
    module, in this case called 'prefs.cmo', and a sample RASCL file
    called 'config.example', which may be used as a default runtime
    configuration file. 

    {3 Using the generated module}

    The module generated by [pa_rascl.cmo] contains a single public
    function that generates a configuration object. Its signature is
    as follows:
    {[
    <function_name> : string list -> <object>
    ]}
    The function name is taken from the [config-name] specified in
    the template. I.e., if the template contains:
    {[
    prefs = {
      ....
    }
    ]}
    then the function name will be [prefs].

    The argument to this function is a list of RASCL configuration
    files. These files will be loaded in the order that they appear in
    the list, and the values they contain will populate the configuration
    object. Later values replace earlier ones. Thus, typically the file
    list will contain a system-wide configuration file, then a
    user-specific file. If defaults have been defined for all properties,
    then the file list may be empty.
    
    RASCL does not handle any errors encountered in trying to read
    configuration files, so your application code must ensure that all
    files exist and are readable.

    The function returns an immediate object (i.e. it is not an instance
    of any class) which inherits from [ConfigObject.root_config]. Thus it
    it provides [#load] and [#dump] as described in the [root_config]
    documentation. Its other public methods correspond to the property
    names in your configuration template.

    {3 Using the configuration object}

    For every property defined in your configuration template, the
    configuration object will have a 0-argument 'getter' method named
    [<prop_name>], and a 1-argument 'setter' named [set_<prop_name>].
    The type returned by the 'getter', and of the argument to the
    'setter', will be whatever you specified in the template. Thus,
    if your template consists of:
    {[
    prefs = {
      font_size : int = 12
      color : string = "black"
    }
    ]}
    then you can access the configuration object in the following manner:
    {[
    prefs_obj#font_size ==> 12  
    prefs_obj#color ==> "black"
    prefs_obj#set_font_size 10 ==> ()
    prefs_obj#set_color "red" ==> ()
    ]}

    Nested dictionaries in the template (and in configuration files)
    correspond to nested sub-objects in the configuration object. Thus,
    given the following template:
    {[
    prefs = {
      font = {
        family : string = "bitstream vera sans"
        size : int = 10
        weight : string = "bold"
        slant : string = "roman"
      }
    }
    ]}
    the object will have the following interface:
    {[
    <
      load : unit -> unit
      dump : unit
      font : <
        dump : unit
        family : string
        set_family : string -> unit
        size : int
        set_size : int -> unit
        weight : string
        set_weight : string -> unit
        slant : string
        set_slant : string -> unit
      >
    >
    ]}
    Note that there is no [#set_font] method. Nested objects cannot
    be added or removed.

    *)
end

val read_config_file : string -> Dict.t
  (** [read_config_file path] parses the configuration file indicated by
      [path] and returns the resulting dictionary. *)
