(* rasclDict.ml -- Copyright  2006 by Matthew C. Gushee 
   This program is free software, released under the terms of the BSD
   license. See the file LICENSE for details. *)

(* ================================================================== *)
(* --  TYPES, EXCEPTIONS, CONSTANTS  -------------------------------- *)
type data = S of string | L of string list | D of t
and t = {
  data : (string, data) Hashtbl.t;
  mutable locked : bool;
}

exception LockViolation
exception Found

let size = ref 53
(* ================================================================== *)


(* ================================================================== *)
(* --  PRIVATE UTILITY FUNCTIONS  ----------------------------------- *)
let rec get_subdict path dict =
  match path with
  | [] -> dict 
  | h::t ->
    ( let sub = Hashtbl.find dict.data h in
      match sub with
      | D dict' -> get_subdict t dict'
      | _ -> failwith ("'" ^ h ^ "' does not represent a subdictionary.") )
   
let rec detach_subdict path dict =
  let head, tail =
    match path with
    | [] -> failwith "Path must have at least one step." 
    | h::t -> h, t in
  let res = Hashtbl.find dict.data head in
  match res, tail with
  | D dict', [] ->
    let result = Hashtbl.copy dict'.data in
    Hashtbl.remove dict.data head;
    { data = result; locked = dict.locked }
  | D dict', path' -> detach_subdict path' dict'
  | _, _ -> failwith ("Path does not represent a subdictionary.")


(* ================================================================== *)

(* ================================================================== *)
(* --  PUBLIC FUNCTIONS  -------------------------------------------- *)

let create ?(locked=false) () =
  { data = Hashtbl.create !size;
    locked = locked }


let get ?prefix key dict =
  let d =
    match prefix with
    | Some path -> get_subdict path dict
    | None -> dict in
  Hashtbl.find d.data key

let get_string ?prefix key dict =
  match (get ?prefix key dict) with
  | S s -> s
  | L _ -> failwith "Type error: expected string, but value is a list."
  | D _ -> failwith "Type error: expected string, but value is a subdictionary."

let get_list ?prefix key dict =
  match (get ?prefix key dict) with
  | L l -> l
  | S _ -> failwith "Type error: expected list, but value is a string."
  | D _ -> failwith "Type error: expected list, but value is a subdictionary."

let get_dict ?prefix key dict =
  match (get ?prefix key dict) with
  | D d -> d
  | S _ -> failwith "Type error: expected dictionary, but value is a string."
  | L _ -> failwith "Type error: expected dictionary, but value is a list."


let set ?prefix key value dict =
  let d =
    match prefix with
    | Some path -> get_subdict path dict
    | None -> dict in
  Hashtbl.replace d.data key value

let set_string ?prefix key value dict = set ?prefix key (S value) dict

let set_list ?prefix key value dict = set ?prefix key (L value) dict

let set_dict ?prefix key value dict = set ?prefix key (D value) dict


let iter f dict = Hashtbl.iter f dict.data

let map f dict =
  let newdict = create ~locked:dict.locked () in
  Hashtbl.iter ( fun k v -> Hashtbl.replace newdict.data k (f k v) ) dict.data;
  newdict

let filter f dict =
  let newdict = create ~locked:dict.locked () in
  Hashtbl.iter
    ( fun k v -> if (f k v) then Hashtbl.replace newdict.data k v )
    dict.data;
  newdict

(* Note that this is a deep-copy operation *)
let rec copy dict =
  let newdict = create ~locked:dict.locked () in
  Hashtbl.iter
    ( fun k v ->
      match v with
      | D sub -> Hashtbl.replace newdict.data k (D (copy sub))
      | _ -> () )
    newdict.data;
  newdict

let do_in context actions dict =
  let d = get_subdict context dict in
  List.iter ( fun act -> act d ) actions

let get_from context keys dict =
  let d = get_subdict context dict in
  List.map ( fun key -> (key, get key d) ) keys

let set_in context kvlist dict =
  let d = get_subdict context dict in
  List.iter ( fun (k, v) -> set k v d ) kvlist
  
(* ================================================================== *)


(* ================================================================== *)
(* --  UNSAFE  ------------------------------------------------------ *)
module Unsafe = struct
  (** The name "Unsafe" does not indicate that it is unsafe to *use*
    * these functions, just that it is unsafe to rely on the interface.
    * Use at your own risk. *)
  let set_size s = size := s
  let get_subdict = get_subdict
  let detach_subdict = detach_subdict
  let alist_of_dict ?(strict=false) ?prefix dict =
    let d =
      match prefix with
      | Some path -> get_subdict path dict
      | None -> dict in
    let extract k v elts =
      match v with
      | S s -> (k, s)::elts 
      | _ when strict -> failwith "Can't extract assoc list."
      | _ -> elts in
    Hashtbl.fold extract d.data []
  let exists ?prefix pred dict =
    let d =
      match prefix with
      | Some path -> get_subdict path dict
      | None -> dict in
    try
      ( Hashtbl.iter
          ( fun k v -> if pred k v then raise Found else () )
          d.data;
        false )
    with Found -> true
  let lock_structure dict = dict.locked <- true
  let unlock_structure dict = dict.locked <- false
  let rec update_from src dest =
    Hashtbl.iter
      ( fun k srcval ->
        let destval =
          try Some (Hashtbl.find dest.data k)
          with Not_found -> None in
        match srcval, destval with
        | D dict, Some (D dict') -> update_from dict dict' 
        | _, Some (D _) when dest.locked -> failwith "structure is locked"
        | D dict, _ when dest.locked -> failwith "structure is locked"
        | _, _ -> Hashtbl.replace dest.data k srcval )
      src.data
  let dump dict =
    let string_of_list l =
      let rec build_string ?(init=false) l' str =
        match l' with
        | [] -> str ^ "]" 
        | h::t when init -> build_string t ("[" ^ h)
        | h::t -> build_string t (str ^ ", " ^ h) in
      build_string ~init:true l "" in
    let rec do_dump dict' indent =
      Hashtbl.iter
        ( fun k v ->
          match v with
          | S s ->
            Printf.printf "%s%s : %s\n" indent k s;
            flush stdout
          | L l ->
            Printf.printf "%s%s : %s\n" indent k (string_of_list l);
            flush stdout
          | D d ->
            Printf.printf "%s%s : {\n" indent k;
            flush stdout;
            do_dump d (indent ^ "  ");
            Printf.printf "%s}\n" indent )
        dict'.data in
    do_dump dict ""
end
(* ================================================================== *)
