(* rasclObject.ml -- Copyright  2006 by Matthew C. Gushee 
   This program is free software, released under the terms of the BSD
   license. See the file LICENSE for details. *)

module RP = RasclParser
module RL = RasclLexer
module RD = RasclDict
open RasclDict

class virtual config_base data defaults prefix = object(self)
  val mutable data_:RD.t = data
  val mutable defaults_:RD.t = defaults
  val mutable prefix_:string list option = prefix

  method private update_data d =
    RD.Unsafe.update_from d data_;
    RD.Unsafe.lock_structure data_
  method private update_defaults d =
    RD.Unsafe.update_from d defaults_;
    RD.Unsafe.lock_structure defaults_


  method private get k =
    try RD.get ?prefix:prefix_ k data_
    with Not_found -> RD.get ?prefix:prefix_ k defaults_ 

  method private set k v =
    RD.set ?prefix:prefix_ k v data_

  method private get_string k =
    match self#get k with
    | S s -> s
    | _ -> failwith "Type error"
  method private get_bool k =
    match self#get k with
    | S s -> bool_of_string s
    | _ -> failwith "Type error"
  method private get_int k =
    match self#get k with
    | S s -> int_of_string s
    | _ -> failwith "Type error"
  method private get_float k =
    match self#get k with
    | S s -> float_of_string s
    | _ -> failwith "Type error"
  method private get_list k =
    match self#get k with
    | L l -> l
    | _ -> failwith "Type error"
  method private get_dict k =
    match self#get k with
    | D d -> d
    | _ -> failwith "Type error"
  method private get_string_list = self#get_list
  method private get_bool_list k =
    List.map bool_of_string (self#get_list k)
  method private get_int_list k =
    List.map int_of_string (self#get_list k)
  method private get_float_list k =
    List.map float_of_string (self#get_list k)
  method private set_string k v = self#set k (S v) 
  method private set_bool k v = self#set k (S (string_of_bool v))
  method private set_int k v = self#set k (S (string_of_int v))
  method private set_float k v = self#set k (S (string_of_float v))
  method private set_list k v = self#set k (L v) 
  method private set_string_list = self#set_list
  method private set_bool_list k v =
    self#set_list k (List.map string_of_bool v)
  method private set_int_list k v =
    self#set_list k (List.map string_of_int v)
  method private set_float_list k v =
    self#set_list k (List.map string_of_float v)

  method dump =
    print_endline
      "DATA *******************************************************************";
    flush stdout;
    RD.Unsafe.dump data_;
    print_endline
      "DEFAULTS ***************************************************************";
    flush stdout;
    RD.Unsafe.dump defaults_
end


class virtual root_config (srcs:string list) (dest:string) data defaults =
object(self)
  inherit config_base data defaults None as super

  method load () =
    let check_file f =
      Sys.file_exists f
    and do_load f =
      let ic = open_in f in
      let buf = Lexing.from_channel ic in
      let newdata = RP.config_dict RL.dict buf in
      super#update_data newdata;
      close_in ic in
    let files = List.filter check_file srcs in
    match files with
    | [] -> failwith "No valid config files available."  
    | _ -> List.iter do_load files
    
end

class virtual sub_config data defaults prefix = object(self)
  inherit config_base data defaults (Some prefix) as super

end
